"""
Module for storing keys in Streamlit session state.
"""

import streamlit as st
import pandas as pd
import copy
import re


ss = st.session_state
newCase = "New Case..."
loadCaseFile = "Upload Case File..."
help1000 = "Value is in \\$1,000 denoted \\$k."


def init():
    """
    Initialize variables through a function as it will only happen once through module.
    """
    # Dictionary of dictionaries for each case.
    global ss
    ss = st.session_state
    if "cases" not in ss:
        ss.cases = {
            newCase: {"iname0": "", "status": "unkown", "caseStatus": "new", "id": 0},
            loadCaseFile: {"iname0": "", "status": "unkown", "caseStatus": "new", "id": 1},
        }

    if "nextid" not in ss:
        ss.nextid = 2

    # Variable for storing name of current case.
    if "currentCase" not in ss:
        ss.currentCase = loadCaseFile
        # ss.currentCase = newCase


init()


def genCaseKey(key):
    """
    This function is to generate global keys that are uniquely associated with a case.
    """
    return f"{getCaseKey('id')}_{key}"


def setCaseId(casename):
    if casename in ss.cases:
        myid = ss.nextid
        ss.cases[casename]["id"] = myid
        ss.nextid = myid + 1
        return myid
    else:
        raise RuntimeError(f"{casename} not found.")


def getKeyInCase(key, casename):
    if casename in ss.cases and key in ss.cases[casename]:
        return ss.cases[casename][key]
    return None


def allCaseNames() -> list:
    return list(ss.cases)


def onlyCaseNames() -> list:
    caseList = list(ss.cases)
    caseList.remove(newCase)
    caseList.remove(loadCaseFile)
    return caseList


def runOncePerSession(func):
    key = "oNcE_" + func.__name__
    if getGlobalKey(key) is None:
        func()
        storeGlobalKey(key, 1)


def runOncePerCase(func):
    key = "oNcE_" + func.__name__
    if key not in ss.cases[ss.currentCase]:
        func()
        storeCaseKey(key, 1)


def refreshCase(adic):
    """
    When a case is duplicated, reset all the runOnce functions.
    """
    for key in list(adic):
        if key.startswith("oNcE_"):
            del adic[key]


def resetTimeLists():
    setCaseKey("stTimeLists", None)
    setCaseKey("timeList0", None)
    if getCaseKey("status") == "married":
        setCaseKey("timeList1", None)


def getIndex(item, choices):
    try:
        i = choices.index(item)
    except ValueError:
        st.error(f"Value {item} not found in {choices}.")
        return None

    return i


def currentCaseName() -> str:
    return ss.currentCase


def switchToCase(key):
    ss.currentCase = ss[key]


def switchToCaseName(casename):
    if casename not in ss.cases:
        raise RuntimeError(f"No such case {casename}.")
    ss.currentCase = casename


def isIncomplete():
    return (
        currentCaseName() == ""
        or getCaseKey("iname0") in [None, ""]
        or (getCaseKey("status") == "married" and getCaseKey("iname1") in [None, ""])
    )


def caseHasNotRun():
    return getCaseKey("caseStatus") not in ["solved", "unsuccessful"]


def isCaseUnsolved():
    if caseHasNoPlan():
        return True
    return getCaseKey("caseStatus") != "solved"


def caseHasNoPlan():
    return getCaseKey("plan") is None


def caseHasPlan():
    return getCaseKey("plan") is not None


def caseIsRunReady():
    return not caseIsNotRunReady() and getCaseKey("caseStatus") in ["modified", "new"]


def caseIsNotRunReady():
    return (
        getCaseKey("plan") is None
        or getCaseKey("objective") is None
        or getCaseKey("rateType") is None
        or getCaseKey("interpMethod") is None
        or getCaseKey("spendingProfile") is None
        or getCaseKey("allocType") is None
    )


def caseIsNotMCReady():
    """
    Check that rates are  set to some stochastic method before MC run.
    """
    return caseIsNotRunReady() or getCaseKey("rateType") != "varying" or "tochastic" not in getCaseKey("varyingType")


def currentCaseDic() -> dict:
    return ss.cases[ss.currentCase]


def setCurrentCase(case):
    if case not in ss.cases:
        raise RuntimeError(f"Case {case} not found in dictionary")
    ss.currentCase = case


def duplicateCase():
    baseName = re.sub(r"\s*\(\d+\)$", "", ss.currentCase)
    for i in range(1, 10):
        dupname = baseName + f" ({i})"
        if dupname not in ss.cases:
            break
    else:
        raise RuntimeError("Exhausted number of duplicates")

    # Copy everything except the plan itself.
    # print(ss.currentCase, "->", ss.cases[ss.currentCase])
    currentPlan = ss.cases[ss.currentCase]["plan"]
    ss.cases[ss.currentCase]["plan"] = None
    ss.cases[dupname] = copy.deepcopy(ss.cases[ss.currentCase])
    ss.cases[ss.currentCase]["plan"] = currentPlan

    ss.cases[dupname]["name"] = dupname
    for key in ["summaryDf", "histoPlot", "histoSummary", "monteCarloPlot", "monteCarloSummary"]:
        ss.cases[dupname][key] = None

    setCaseId(dupname)
    ss.cases[dupname]["duplicate"] = True
    refreshCase(ss.cases[dupname])
    ss.currentCase = dupname
    st.toast("Case duplicated.")


def createCaseFromFile(strio):
    import owlbridge as owb

    name, dic = owb.createCaseFromFile(strio)
    if name == "":
        return False
    elif name in ss.cases:
        st.error(f"Case name '{name}' already exists.")
        return False

    ss.cases[name] = dic
    setCaseId(name)
    setCurrentCase(name)
    return True


def createNewCase(case):
    if case != "newcase":
        st.error(f"Expected 'newcase' but got '{case}'.")
        return

    # Widget stored case name in _newcase.
    casename = ss._newcase

    if casename == "":
        return

    if casename in ss.cases:
        st.error(f"Case name '{casename}' already exists.")
        return

    ss.cases[casename] = {"name": casename, "caseStatus": "unknown", "logs": None}
    setCaseId(casename)
    setCurrentCase(casename)


def renameCase(key):
    if ss.currentCase == newCase or ss.currentCase == loadCaseFile:
        return
    newname = ss[key]
    if newname in ss.cases:
        st.error(f"Case name '{newname}' already exists.")
        return

    plan = getCaseKey("plan")
    ss.cases[newname] = ss.cases.pop(ss.currentCase)
    ss.cases[newname]["name"] = newname
    if plan:
        plan.rename(newname)
        ss.cases[newname]["caseStatus"] = "modified"
    setCurrentCase(newname)


def deleteCurrentCase():
    if ss.currentCase == newCase or ss.currentCase == loadCaseFile:
        return
    del ss.cases[ss.currentCase]
    setCurrentCase(loadCaseFile)


def dumpSession():
    print("State Dump:", ss)


def dumpCase(case=None):
    if case is None:
        case = ss.currentCase
    print("Case Dump:", ss.cases[case])


def setpull(key):
    return setCaseKey(key, ss[genCaseKey(key)])


def storepull(key):
    return storeCaseKey(key, ss[genCaseKey(key)])


def setCaseKey(key, val):
    ss.cases[ss.currentCase][key] = val
    flagModified()
    return val


def flagModified():
    ss.cases[ss.currentCase]["caseStatus"] = "modified"
    ss.cases[ss.currentCase]["summaryDf"] = None


def storeCaseKey(key, val):
    ss.cases[ss.currentCase][key] = val
    return val


def initCaseKey(key, val):
    """
    Only set the case local key if unset.
    """
    if key not in ss.cases[ss.currentCase]:
        setCaseKey(key, val)


def initGlobalKey(key, val):
    if key not in ss:
        ss[key] = val


def getCaseKey(key):
    return ss.cases[ss.currentCase].get(key)


def storeGlobalKey(key, val):
    ss[key] = val
    return val


def getGlobalKey(key):
    return ss.get(key)


def getDict(key=ss.currentCase):
    return ss.cases[key]


def getAccountBalances(ni):
    bal = [[], [], []]
    accounts = ["txbl", "txDef", "txFree"]
    for j, acc in enumerate(accounts):
        for i in range(ni):
            bal[j].append(getCaseKey(acc + str(i)))

    return bal


def colorBySign(val):
    color = "green" if "\u2191" in val else "red" if "\u2193" in val else None

    return f"color:{color};" if color else ""


def compareSummaries():
    df = getCaseKey("summaryDf")
    if df is None:
        return None
    for case in onlyCaseNames():
        if case == currentCaseName():
            continue
        odf = ss.cases[case]["summaryDf"]
        if odf is None or set(odf.columns) != set(df.columns):
            continue
        df = pd.concat([df, odf])

    if df.shape[0] > 1:
        # Unroll to subtract $tring representation of numbers.
        for col in range(1, df.shape[1] - 5):
            strval = df.iloc[0, col]
            if isinstance(strval, str) and strval[0] == "$":
                f0val = float(strval[1:].replace(",", ""))
                for row in range(1, df.shape[0]):
                    fnval = float(df.iloc[row, col][1:].replace(",", ""))
                    diff = fnval - f0val
                    sign = "\u2191" if diff > 0 else "\u2193" if diff < 0 else "\u2192"
                    df.iloc[row, col] = f"{sign} ${abs(diff):,.0f}"

    return df.transpose()


def getSolveParameters():
    maximize = getCaseKey("objective")
    if maximize is None:
        return None
    if "spending" in maximize:
        objective = "maxSpending"
    else:
        objective = "maxBequest"

    options = {}
    optList = ["netSpending", "maxRothConversion", "noRothConversions",
               "startRothConversions", "withMedicare", "bequest", "solver",
               "spendingSlack", "oppCostX", "xorConstraints", "withSCLoop",]
    for opt in optList:
        val = getCaseKey(opt)
        if val is not None:
            options[opt] = val

    if getCaseKey("readRothX"):
        options["maxRothConversion"] = "file"

    previousMAGIs = getPreviousMAGIs()
    if previousMAGIs[0] > 0 or previousMAGIs[1] > 0:
        options["previousMAGIs"] = previousMAGIs

    return objective, options


def getIndividualAllocationRatios():
    generic = []
    ni = 2 if getCaseKey("status") == "married" else 1
    for i in range(ni):
        initial = []
        final = []
        for k1 in range(4):
            initial.append(int(getCaseKey(f"j3_init%{k1}_{i}")))
            final.append(int(getCaseKey(f"j3_fin%{k1}_{i}")))
        gen = [initial, final]
        generic.append(gen)

    return generic


def getAccountAllocationRatios():
    accounts = [[], [], []]
    ni = 2 if getCaseKey("status") == "married" else 1
    for i in range(ni):
        for j1 in range(3):
            initial = []
            final = []
            for k1 in range(4):
                initial.append(int(getCaseKey(f"j{j1}_init%{k1}_{i}")))
                final.append(int(getCaseKey(f"j{j1}_fin%{k1}_{i}")))
            tmp = [initial, final]
            accounts[j1].append(tmp)

    return accounts


def getPreviousMAGIs():
    backMAGIs = [0., 0.]
    for ii in range(2):
        val = getCaseKey(f"MAGI{ii}")
        if val:
            backMAGIs[ii] = float(val)

    return backMAGIs


def getFixedIncome(ni, what):
    amounts = []
    ages = []
    indexed = []
    for i in range(ni):
        amounts.append(getCaseKey(what + "Amt" + str(i)))
        ages.append(getCaseKey(what + "Age" + str(i)))
        if what == "p":
            indexed.append(getCaseKey(what + "Idx" + str(i)))

    return amounts, ages, indexed


def getIntNum(text, nkey, disabled=False, callback=setpull, step=1, help=None, min_value=0, max_value=None):
    return st.number_input(
        text,
        value=int(getCaseKey(nkey)),
        disabled=disabled,
        min_value=min_value,
        max_value=max_value,
        step=step,
        help=help,
        on_change=callback,
        args=[nkey],
        key=genCaseKey(nkey),
    )


def getNum(text, nkey, disabled=False, callback=setpull, step=10.0, min_value=0.0,
           max_value=None, format="%.1f", help=None):
    return st.number_input(
        text,
        value=float(getCaseKey(nkey)),
        disabled=disabled,
        step=step,
        help=help,
        min_value=min_value,
        max_value=max_value,
        format=format,
        on_change=callback,
        args=[nkey],
        key=genCaseKey(nkey),
    )


def getText(text, nkey, disabled=False, callback=setpull, placeholder=None, help=None):
    return st.text_input(
        text,
        value=getCaseKey(nkey),
        disabled=disabled,
        on_change=callback,
        args=[nkey],
        key=genCaseKey(nkey),
        placeholder=placeholder,
        help=help,
    )


def getLongText(text, nkey, disabled=False, callback=setpull, placeholder=None, help=None):
    return st.text_area(
        text,
        value=getCaseKey(nkey),
        disabled=disabled,
        on_change=callback,
        args=[nkey],
        key=genCaseKey(nkey),
        placeholder=placeholder,
        help=help,
    )


def getRadio(text, choices, nkey, callback=setpull, disabled=False, help=None):
    try:
        index = choices.index(getCaseKey(nkey))
    except ValueError:
        st.error(f"Value '{getCaseKey(nkey)}' not available. Defaulting to '{choices[0]}'.")
        setCaseKey(nkey, choices[0])
        index = 0

    return st.radio(
        text,
        choices,
        index=index,
        on_change=callback,
        args=[nkey],
        key=genCaseKey(nkey),
        disabled=disabled,
        horizontal=True,
        help=help,
    )


def getToggle(text, nkey, callback=setpull, disabled=False, help=None):
    return st.toggle(
        text,
        value=getCaseKey(nkey),
        on_change=callback, args=[nkey], disabled=disabled,
        key=genCaseKey(nkey), help=help
    )


def divider(color, width="auto"):
    st.html("<style> hr {border-color: %s;width: %s}</style><hr>" % (color, width))


def getColors():
    def_theme = "dark"
    bc = "#0E1117" if def_theme == "dark" else "#FFFFFF"
    fc = "#FAFAFA" if def_theme == "dark" else "#31333F"

    return bc, fc


def flagCurrentCase(caseName):
    if caseName == currentCaseName():
        return caseName + "🔻"
    else:
        return caseName


def titleBar(txt, allCases=False):
    if allCases:
        choices = allCaseNames()
        helpmsg = "Select an existing case, or create a new one from scratch or from a *case* parameter file."
    else:
        choices = onlyCaseNames()
        helpmsg = "Select an existing case."

    header = st.container()
    # header.title("Here is a sticky header")
    header.write("""<div class='fixed-header'/>""", unsafe_allow_html=True)

    bc, fc = getColors()

    position = getGlobalKey("position")
    # Custom CSS for the sticky header
    st.markdown(
        f"""<style>
    div[data-testid="stVerticalBlock"] div:has(div.fixed-header) {{
        border-radius: 10px;
        position: {position};
        background: linear-gradient(to right, #551b1b, #EFB761);
        /* background-color: %s; */
        color: %s;
        top: 2.875rem;
        z-index: 100;
    }}
</style>""" % (bc, fc), unsafe_allow_html=True
    )
    with header:
        col1, col2, col3, col4 = st.columns([0.005, 0.6, 0.4, 0.01], gap="small")
        with col2:
            st.write("## " + txt)
        with col3:
            nkey = txt
            ret = st.selectbox(
                "Case selector",
                choices,
                help=helpmsg,
                index=getIndex(currentCaseName(), choices),
                format_func=flagCurrentCase,
                key=nkey,
                on_change=switchToCase,
                args=[nkey],
            )

        divider("white", "99%")

    return ret




import owlplanner as owl


def getWaC(exdir, base):
    import os
    wac = base.replace('case_', '')
    wac = wac.replace('-spending', '')
    wac = wac.replace('-bequest', '')
    wac = exdir + wac + '.xlsx'
    if os.path.exists(wac):
        return wac
    else:
        return ''


def test_allcases():
    exdir = './examples/'
    for case in ['case_john+sally',
                 'case_jack+jill',
                 'case_joe',
                 'case_kim+sam-spending',
                 'case_kim+sam-bequest']:
        p = owl.readConfig(exdir + case)
        wac = getWaC(exdir, case)
        if wac != '':
            p.readContributions(wac)
        p.resolve()


def test_historical():
    exdir = './examples/'
    case = 'case_jack+jill'
    p = owl.readConfig(exdir + case)
    wac = getWaC(exdir, case)
    if wac != '':
        p.readContributions(wac)
    options = p.solverOptions
    objective = p.objective
    p.runHistoricalRange(objective, options, 1969, 2023)


def test_MC():
    exdir = './examples/'
    case = 'case_jack+jill'
    p = owl.readConfig(exdir + case)
    wac = getWaC(exdir, case)
    if wac != '':
        p.readContributions(wac)
    options = p.solverOptions
    objective = p.objective
    p.runMC(objective, options, 20)



import streamlit as st

import sskeys as kz
import owlbridge as owb


FXRATES = {
    "conservative": [7, 4, 3.3, 2.8],
    "optimistic": [10, 6, 5, 3],
    "historical average": [0, 0, 0, 0],
    "user": [7, 4, 3.3, 2.8],
}

rateChoices = ["fixed", "varying"]
fixedChoices = list(FXRATES)
varyingChoices = ["historical", "histochastic", "stochastic"]


def updateFixedRates(key, pull=True):
    if pull:
        fxType = kz.setpull(key)
    else:
        fxType = key

    rates = FXRATES[fxType]
    for j in range(4):
        kz.setCaseKey("fxRate" + str(j), rates[j])
    owb.setRates()


def updateRates(key):
    kz.setpull(key)
    if kz.getCaseKey(key) == "fixed":
        updateFixedRates(kz.getCaseKey("fixedType"), False)
    else:
        owb.setRates()


def initRates():
    if kz.getCaseKey("rateType") == rateChoices[0] and kz.getCaseKey("fixedType") == fixedChoices[0]:
        updateFixedRates(fixedChoices[0], False)
    else:
        owb.setRates()
    kz.flagModified()


kz.initCaseKey("rateType", rateChoices[0])
kz.initCaseKey("fixedType", fixedChoices[0])
kz.initCaseKey("varyingType", varyingChoices[0])

ret = kz.titleBar(":material/monitoring: Rates Selection")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    kz.runOncePerCase(initRates)
    kz.initCaseKey("yfrm", owb.FROM)
    kz.initCaseKey("yto", owb.TO)
    helpmsgSP500 = """Rate also includes dividends.
Unless historical, S&P 500 can represent any mix of equities
(domestic, international, emerging, ...).
"""
    helpFixed = """A 2025 roundup of expert opinions on stock and bond return
forecasts for the next decade can be found
[here](https://www.morningstar.com/portfolios/experts-forecast-stock-bond-returns-2025-edition)."""
    helpCash = """Here, "Cash Assets" are TIPS-like securities assumed to track inflation."""

    st.write("#### :orange[Type of Rates]")
    col1, col2 = st.columns(2, gap="large", vertical_alignment="top")
    with col1:
        kz.getRadio("## Annual rates type", rateChoices, "rateType", updateRates)

    if kz.getCaseKey("rateType") == "fixed":
        with col2:
            fxType = kz.getRadio("Select fixed rates", fixedChoices, "fixedType", updateFixedRates,
                                 help=helpFixed)

        st.divider()
        ro = fxType != "user"

        st.write("#### :orange[Fixed Rate Values (%)]")
        rates = FXRATES[fxType]
        for j in range(4):
            kz.initCaseKey("fxRate" + str(j), rates[j])

        col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="top")
        with col1:
            kz.getNum("S&P 500", "fxRate0", ro, step=1.0, help=helpmsgSP500, callback=updateRates)

        with col2:
            kz.getNum("Corporate Bonds Baa", "fxRate1", ro, step=1.0, callback=updateRates)

        with col3:
            kz.getNum("10-y Treasury Notes", "fxRate2", ro, step=1.0, callback=updateRates)

        with col4:
            kz.getNum("Cash Assets/Inflation", "fxRate3", ro, step=1.0, help=helpCash, callback=updateRates)

    elif kz.getCaseKey("rateType") == "varying":
        with col2:
            kz.getRadio("Select varying rates", varyingChoices, "varyingType", callback=updateRates)

    else:
        st.error("Logic error")

    if (kz.getCaseKey("rateType") == "fixed" and "hist" in kz.getCaseKey("fixedType")) or (
        kz.getCaseKey("rateType") == "varying" and "hist" in kz.getCaseKey("varyingType")
    ):

        col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="top")
        with col3:
            maxValue = owb.TO if kz.getCaseKey("varyingType") == "historical" else kz.getCaseKey("yto") - 1
            st.number_input(
                "Starting year",
                min_value=owb.FROM,
                max_value=maxValue,
                value=kz.getCaseKey("yfrm"),
                on_change=updateRates,
                args=["yfrm"],
                key=kz.genCaseKey("yfrm"),
            )

        with col4:
            ishistorical = kz.getCaseKey("rateType") == "varying" and kz.getCaseKey("varyingType") == "historical"
            st.number_input(
                "Ending year",
                max_value=owb.TO,
                min_value=kz.getCaseKey("yfrm") + 1,
                value=kz.getCaseKey("yto"),
                disabled=ishistorical,
                on_change=updateRates,
                args=["yto"],
                key=kz.genCaseKey("yto"),
            )

    if kz.getCaseKey("rateType") == "varying":
        st.divider()
        st.write("#### :orange[Stochastic Parameters]")
        ro = kz.getCaseKey("varyingType") != "stochastic"
        st.write("##### Means (%)")
        col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="top")
        with col1:
            kz.initCaseKey("mean0", 0)
            kz.getNum("S&P 500", "mean0", ro, step=1.0, help=helpmsgSP500, min_value=-9.0, callback=updateRates)

        with col2:
            kz.initCaseKey("mean1", 0)
            kz.getNum("Corporate Bonds Baa", "mean1", ro, step=1.0, min_value=-9.0, callback=updateRates)

        with col3:
            kz.initCaseKey("mean2", 0)
            kz.getNum("10-y Treasury Notes", "mean2", ro, step=1.0, min_value=-9.0, callback=updateRates)

        with col4:
            kz.initCaseKey("mean3", 0)
            kz.getNum("Cash Assets/Inflation", "mean3", ro, step=1.0, min_value=-9.0, callback=updateRates)

        st.write("##### Volatility (%)")
        col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="top")
        with col1:
            kz.initCaseKey("stdev0", 0)
            kz.getNum("S&P 500", "stdev0", ro, step=1.0, callback=updateRates)

        with col2:
            kz.initCaseKey("stdev1", 0)
            kz.getNum("Corporate Bonds Baa", "stdev1", ro, step=1.0, callback=updateRates)

        with col3:
            kz.initCaseKey("stdev2", 0)
            kz.getNum("10-y Treasury Notes", "stdev2", ro, step=1.0, callback=updateRates)

        with col4:
            kz.initCaseKey("stdev3", 0)
            kz.getNum("Cash Assets/Inflation", "stdev3", ro, step=1.0, callback=updateRates)

        st.write("##### Correlation matrix")
        col1, col2, col3, col4 = st.columns(4, gap="large", vertical_alignment="top")
        with col1:
            kz.initCaseKey("diag1", 1)
            kz.getNum("S&P 500", "diag1", True, format="%.2f", callback=None)

        with col2:
            kz.initCaseKey("corr1", 0.0)
            kz.getNum("(1,2)", "corr1", ro, step=0.1, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=updateRates)
            kz.initCaseKey("diag2", 1.0)
            kz.getNum("Corporate Bonds Baa", "diag2", True, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=None)

        with col3:
            kz.initCaseKey("corr2", 0.0)
            kz.getNum("(1,3)", "corr2", ro, step=0.1, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=updateRates)
            kz.initCaseKey("corr4", 0.0)
            kz.getNum("(2,3)", "corr4", ro, step=0.1, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=updateRates)
            kz.initCaseKey("diag3", 1.0)
            kz.getNum("10-y Treasury Notes", "diag3", True, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=None)

        with col4:
            kz.initCaseKey("corr3", 0.0)
            kz.getNum("(1,4)", "corr3", ro, step=0.1, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=updateRates)
            kz.initCaseKey("corr5", 0.0)
            kz.getNum("(2,4)", "corr5", ro, step=0.1, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=updateRates)
            kz.initCaseKey("corr6", 0.0)
            kz.getNum("(3,4)", "corr6", ro, step=0.1, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=updateRates)
            kz.initCaseKey("diag4", 1.0)
            kz.getNum("Cash Assets/Inflation", "diag4", True, format="%.2f",
                      min_value=-1.0, max_value=1.0, callback=None)

    st.divider()
    if kz.getCaseKey("rateType") == "varying":
        col1, col2 = st.columns(2, gap="medium")
        owb.showRatesCorrelations(col2)
    else:
        col1, col2 = st.columns([0.6, 0.4], gap="medium")

    owb.showRates(col1)

    # st.divider()
    st.write("#### :orange[Other Rates]")
    col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
    with col1:
        kz.initCaseKey("divRate", 1.8)
        helpmsg = "Average annual (qualified) dividend return rate on stock portfolio for income tax purposes."
        ret = kz.getNum("Dividend rate (%)", "divRate", max_value=100.0, format="%.2f", help=helpmsg, step=1.0)

    st.write("#####")
    st.write("#### :orange[Income taxes]")
    col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
    with col1:
        kz.initCaseKey("heirsTx", 30)
        helpmsg = "Marginal tax rate that heirs would have to pay on inherited tax-deferred balance."
        ret = kz.getNum("Heirs marginal tax rate (%)", "heirsTx", max_value=100.0, help=helpmsg, step=1.0)

    with col2:
        kz.initCaseKey("yOBBBA", 2032)
        helpmsg = "Year at which the OBBBA tax rates are speculated to be expired and return to pre-TCJA rates."
        ret = kz.getIntNum("OBBBA expiration year", "yOBBBA", help=helpmsg)



"""

Owl/utils

This file contains functions for handling data.

Copyright &copy; 2024 - Martin-D. Lacasse

Disclaimers: This code is for educational purposes only and does not constitute financial advice.

"""

######################################################################
import numpy as np


def d(value, f=0, latex=False) -> str:
    """
    Return a string formatting value in $ currency.
    Number of decimals controlled by `f` which defaults to 0.
    """
    if np.isnan(value):
        return "NaN"

    if latex:
        mystr = "\\${:,." + str(f) + "f}"
    else:
        mystr = "${:,." + str(f) + "f}"

    return mystr.format(value)


def pc(value, f=1, mul=100) -> str:
    """
    Return a string formatting decimal value in percent.
    Number of decimals of percent controlled by `f` which defaults to 1.
    """
    mystr = "{:." + str(f) + "f}%"

    return mystr.format(mul * value)


def rescale(vals, fac):
    """
    Rescale all elements of a list or a NumPy array by factor fac.
    """
    if isinstance(vals, (float, int)) or isinstance(vals, np.ndarray):
        return vals * fac
    else:
        for i in range(len(vals)):
            vals[i] *= fac

    return vals


def getUnits(units) -> int:
    """
    Translate multiplication factor for units as expressed by an abbreviation
    expressed in a string. Returns an integer.
    """
    if units is None or units == 1 or units == "1" or units == "one":
        fac = 1
    elif units in {"k", "K"}:
        fac = 1000
    elif units in {"m", "M"}:
        fac = 1000000
    else:
        raise ValueError(f"Unknown units {units}.")

    return fac


# Next two functins could be a one-line lambda functions.
# e.g., krond = lambda a, b: 1 if a == b else 0
def krond(a, b) -> int:
    """
    Kronecker integer delta function.
    """
    return 1 if a == b else 0


def heavyside(x) -> int:
    """
    Heavyside step function.
    """
    return 1 if x >= 0 else 0


def roundCents(values, decimals=2):
    """
    Round values in NumPy array down to second decimal.
    Using fix which is floor towards zero.
    Default is to round to cents (decimals = 2).
    """
    multiplier = 10**decimals

    newvalues = values * multiplier + 0.5 * np.sign(values)

    arr = np.fix(newvalues) / multiplier
    # Remove negative zero-like values.
    arr = np.where((-0.009 < arr) & (arr <= 0), 0, arr)

    return arr



import requests
import streamlit as st
from io import StringIO, BytesIO


cases = ["jack+jill", "joe", "john+sally", "jon+jane", "kim+sam-bequest",
         "kim+sam-spending", "drawdowncalc-comparison-1"]


wages = ["jack+jill", "joe", "john+sally", "jon+jane",
         "kim+sam", "charles"]


def loadCaseExample(case):
    url = f"https://github.com/mdlacasse/owl/blob/main/examples/case_{case}.toml?raw=true"
    response = requests.get(url)
    if response.ok:
        return StringIO(response.text)
    else:
        st.error(f"Failed to load case parameter file from GitHub: {response.status_code}.")
        return None


def loadWagesExample(case):
    url = f"https://github.com/mdlacasse/owl/blob/main/examples/{case}.xlsx?raw=true"
    response = requests.get(url)
    if response.ok:
        return BytesIO(response.content)
    else:
        st.error(f"Failed to load Wages and Contributions file from GitHub: {response.status_code}.")
        return None



import streamlit as st

import sskeys as kz

# Pick one for narrow or wide graphs. That can also be changed in upper-right settings menu.
st.set_page_config(layout="wide", page_title="Owl Retirement Planner")
# st.set_page_config(layout="centered", page_title="Owl Retirement Planner")

kz.init()

st.logo("https://github.com/mdlacasse/Owl/blob/main/docs/images/owl.png?raw=true", size="large")

pages = {
    "Case Setup": [
        st.Page("Create_Case.py", icon=":material/person_add:"),
        st.Page("Wages_and_Contributions.py", icon=":material/work_history:"),
        st.Page("Fixed_Income.py", icon=":material/currency_exchange:"),
        st.Page("Current_Assets.py", icon=":material/savings:"),
        st.Page("Asset_Allocation.py", icon=":material/percent:"),
        st.Page("Rates_Selection.py", icon=":material/monitoring:"),
        st.Page("Optimization_Parameters.py", icon=":material/tune:"),
    ],
    "Single Scenario": [
        st.Page("Graphs.py", icon=":material/stacked_line_chart:"),
        st.Page("Worksheets.py", icon=":material/data_table:"),
        st.Page("Output_Files.py", icon=":material/description:"),
    ],
    "Multiple Scenarios": [
        st.Page("Historical_Range.py", icon=":material/history:"),
        st.Page("Monte_Carlo.py", icon=":material/finance:"),
    ],
    "Resources": [
        st.Page("Quick_Start.py", icon=":material/rocket_launch:", default=True),
        st.Page("Documentation.py", icon=":material/help:"),
        st.Page("Settings.py", icon=":material/settings:"),
        st.Page("Logs.py", icon=":material/error:"),
        st.Page("About_Owl.py", icon=":material/info:"),
    ],
}

kz.initGlobalKey("menuLocation", "top")
kz.initGlobalKey("position", "sticky")

pg = st.navigation(pages, position=kz.getGlobalKey("menuLocation"))

pg.run()



import streamlit as st

import sskeys as kz
import owlbridge as owb
import tomlexamples as tomlex


def loadWCExample(file):
    if file:
        mybytesio = tomlex.loadWagesExample(file)
        owb.readContributions(mybytesio, file=file)


ret = kz.titleBar(":material/work_history: Wages and Contributions")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    if kz.getCaseKey("timeList0") is None:
        kz.runOncePerCase(owb.resetTimeLists)
    kz.initCaseKey("stTimeLists", None)
    n = 2 if kz.getCaseKey("status") == "married" else 1

    if kz.getCaseKey("stTimeLists") is None:
        original = kz.getCaseKey("timeListsFileName")
        if original is None or original == "None":
            st.info(
                f"Case *'{kz.currentCaseName()}'* makes no reference to a wages and contributions file.\n\n"
                "You can build your own file by directly filling the table(s) below. "
                "Once a case has been successfully run, values can be saved on the **Output Files** page. "
                "Alternatively, you can start from this Excel "
                "[template](https://github.com/mdlacasse/Owl/blob/main/examples/template.xlsx?raw=true) "
                "and upload the file using the widget below."
            )
        elif original != "edited values":
            st.info(
                f"Case *'{kz.currentCaseName()}'* refers to Wages and Contributions file *'{original}'*"
                " that has not yet been uploaded."
            )

    col1, col2 = st.columns(2, gap="large")
    with col1:
        st.write("#### :orange[Upload a *Wages and Contributions* File]")
        kz.initCaseKey("_xlsx", 0)
        stTimeLists = st.file_uploader(
            "Upload values from a Wages and Contributions file...",
            key="_stTimeLists" + str(kz.getCaseKey("_xlsx")),
            type=["xlsx", "ods"],
        )
        if stTimeLists is not None:
            if owb.readContributions(stTimeLists):
                kz.setCaseKey("stTimeLists", stTimeLists)
                # Change key to reset uploader.
                kz.storeCaseKey("_xlsx", kz.getCaseKey("_xlsx") + 1)
                st.rerun()
    with col2:
        tomlexcase = kz.getCaseKey("tomlexcase")
        if tomlexcase in tomlex.wages:
            st.write("#### :orange[Load Example File]")
            st.write("Read associated Wages and Contributions file.")
            helpmsg = "Load associated Wages and Contributions file from GitHub"
            st.button("Load file from GitHub", help=helpmsg,
                      on_click=loadWCExample, args=[tomlexcase])

    st.divider()
    for i in range(n):
        st.write("#### :orange[" + kz.getCaseKey("iname" + str(i)) + "'s Timetable]")
        st.write("""Previous 5 years are used to input past contributions and conversions to Roth accounts.
 This information is needed to enforce the 5-year maturation rule in Roth savings accounts.""")
        df = kz.getCaseKey("timeList" + str(i))
        formatdic = {"year": st.column_config.NumberColumn(None, format="%d", disabled=True)}
        cols = list(df.columns)
        for col in cols[1:-1]:
            formatdic[col] = st.column_config.NumberColumn(None, min_value=0.0, format="accounting")
        formatdic[cols[-1]] = st.column_config.NumberColumn(None, format="accounting")

        newdf = st.data_editor(
            df,
            column_config=formatdic,
            hide_index=True,
            key=kz.currentCaseName() + "_wages" + str(i),
        )
        st.caption("Values are in nominal $.")
        newdf.fillna(0, inplace=True)
        kz.storeCaseKey("_timeList" + str(i), newdf)

        if not df.equals(newdf):
            kz.setCaseKey("timeList" + str(i), newdf)
            st.rerun()

    st.button("Reset to zero", help="Reset all values to zero.", on_click=owb.resetTimeLists)




import ui.sskeys as sskeys


def test_getIndex_found():
    choices = ['a', 'b', 'c']
    assert sskeys.getIndex('b', choices) == 1


def test_getIndex_not_found():
    choices = ['a', 'b', 'c']
    assert sskeys.getIndex('z', choices) is None


def test_caseHasNoPlan(monkeypatch):
    # Mock getKey to return None
    monkeypatch.setattr(sskeys, "getCaseKey", lambda key: None)
    assert sskeys.caseHasNoPlan() is True


def test_caseHasPlan(monkeypatch):
    # Mock getKey to return a non-None value
    monkeypatch.setattr(sskeys, "getCaseKey", lambda key: "something")
    assert sskeys.caseHasNoPlan() is False




from owlplanner import utils as u


def test_rounding():
    number = 1500000.
    number = u.roundCents(number)
    assert number == number
    number = -1500000.
    number = u.roundCents(number)
    assert number == number



import streamlit as st

import sskeys as kz
import owlbridge as owb


st.write("# :material/info: About Owl 🦉")
kz.divider("orange")

st.write(f"This is Owl version {owb.version()} running on Streamlit {st.__version__}.")
# st.balloons()

st.write(
    """
- Owl is an open-source retirement financial planner capable of optimization through linear programming.
- Source code is available from a repository on [GitHub](https://github.com/mdlacasse/owl).
- Mathematical formulation of the linear optimization problem can be
found in this [document](https://github.com/mdlacasse/Owl/blob/main/docs/owl.pdf?raw=true).

Copyright &copy; 2025 - Martin-D. Lacasse

#### :orange[Credits]
- Historical rates are from [Aswath Damodaran](https://pages.stern.nyu.edu/~adamodar/).
- Linear programming optimization solvers are from
[HiGHS](https://highs.dev) and [PuLP](https://coin-or.github.io/pulp/).
It can also run on [MOSEK](https://mosek.com) if available on your computer.
- Owl planner relies on the following [Python](https://python.org) packages:
    - [highspy](https://highs.dev),
    [Matplotlib](https://matplotlib.org),
    [Numpy](https://numpy.org),
    [odfpy](https://https://pypi.org/project/odfpy),
    [openpyxl](https://openpyxl.readthedocs.io),
    [Pandas](https://pandas.pydata.org),
    [Plotly](https://plotly.com),
    [PuLP](https://coin-or.github.io/pulp),
    [Scipy](https://scipy.org),
    [Seaborn](https://seaborn.pydata.org),
    [toml](https://toml.io),
 and [Streamlit](https://streamlit.io) for the front-end.
- Contributors: Josh Williams (noimjosh) for Docker image code,
 kg333 for fixing an error in Docker's instructions,
 Dale Seng (sengsational) for great insights and suggestions,
 Robert E. Anderson (NH-RedAnt) for bug fixes and suggestions,
 Clark Jefcoat (hubcity) for fruitful interactions,
 Benjamin Quinn (blquinn) and Gene Wood (gene1wood) for improvements and for bug fixes.
- Owl image is from [freepik](https://freepik.com).

#### :orange[Bugs and Feature Requests]
Please submit bugs and feature requests through
[GitHub](https://github.com/mdlacasse/owl/issues) if you have a GitHub account
or directly by [email](mailto:martin.d.lacasse@gmail.com).
Or just drop me a line to report your experience with the tool. &#x1F44D;

#### :orange[Privacy]
This app does not store or forward any information. All data entered is lost
after a session is closed. However, you can choose to download selected parts of your
own data to your computer before closing the session. These data will be stored strictly on
your computer and can be used to reproduce a case at a later time.

#### :orange[License]
This software is released under the
[Gnu General Public License v3](https://www.gnu.org/licenses/gpl-3.0.html#license-text).

#### :orange[Disclaimer]
This program is for educatonal purposes only and does not constitute financial advice.

"""
)



import streamlit as st

import sskeys as kz


def getIntInput(i, key, thing, defval=0):
    nkey = key + str(i)
    kz.initCaseKey(nkey, defval)
    inamex = kz.getCaseKey("iname" + str(i))
    st.number_input(
        f"{inamex}'s {thing}", min_value=0, value=kz.getCaseKey(nkey),
        on_change=kz.setpull, args=[nkey], key=kz.genCaseKey(nkey),
    )


def getFloatInput(i, key, thing, defval=0.0):
    nkey = key + str(i)
    kz.initCaseKey(nkey, defval)
    inamex = kz.getCaseKey("iname" + str(i))
    st.number_input(
        f"{inamex}'s {thing}",
        min_value=0.0,
        help=kz.help1000,
        value=float(kz.getCaseKey(nkey)),
        format="%.1f",
        step=10.0,
        on_change=kz.setpull,
        args=[nkey],
        key=kz.genCaseKey(nkey),
    )


def getToggleInput(i, key, thing):
    nkey = key + str(i)
    kz.initCaseKey(nkey, False)
    defval = kz.getCaseKey(nkey)
    st.toggle(thing, on_change=kz.setpull, value=defval, args=[nkey], key=kz.genCaseKey(nkey))


ret = kz.titleBar(":material/currency_exchange: Fixed Income")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    st.write("#### :orange[Social Security]")
    col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
    with col1:
        getFloatInput(0, "ssAmt", "social security annual amount (\\$k)")
        getIntInput(0, "ssAge", "social security age", 67)

    with col2:
        if kz.getCaseKey("status") == "married":
            getFloatInput(1, "ssAmt", "social security annual amount (\\$k)")
            getIntInput(1, "ssAge", "social security age", 67)

    st.divider()
    st.write("#### :orange[Pension]")
    col1, col2, col3 = st.columns(3, gap="large", vertical_alignment="top")
    with col1:
        getFloatInput(0, "pAmt", "pension annual amount (\\$k)")
        getIntInput(0, "pAge", "pension age", 65)
        getToggleInput(0, "pIdx", "Inflafion adjusted")

    with col2:
        if kz.getCaseKey("status") == "married":
            getFloatInput(1, "pAmt", "pension annual amount (\\$k)")
            getIntInput(1, "pAge", "pension age", 65)
            getToggleInput(1, "pIdx", "Inflafion adjusted")



import streamlit as st

import sskeys as kz
import owlbridge as owb

ret = kz.titleBar(":material/description: Output Files")

if ret is None or kz.caseHasNoPlan():
    st.info("Case(s) must be first created before running this page.")
else:
    if kz.caseIsRunReady():
        owb.runPlan()
    elif kz.caseHasNotRun():
        st.info("Case definition is not yet complete. Please visit all pages in *Case Setup*.")

    if kz.isCaseUnsolved():
        st.info("Case status is currently '%s'." % kz.getCaseKey("caseStatus"))
    else:
        caseName = kz.getCaseKey("name")
        df = kz.compareSummaries()
        if df is not None:
            st.write("#### :orange[Synopsis]\n"
                     "This table provides a summary of the current case and"
                     " compares it with other similar cases that ran successfully.")
            styledDf = df[1:].style.map(kz.colorBySign)
            st.dataframe(styledDf, width="stretch")
            st.caption("Values with [legend] are nominal, otherwise in today's \\$. "
                       "Lines starting with » indicate itemized subtotals.")
            col1, col2 = st.columns(2, gap="large")
            helpmsg = "Download synopsis and comparisons as a text file."
            col1.download_button("Download Synopsis", data=df[1:].to_string(),
                                 file_name=f"Synopsis_{caseName}.txt", help=helpmsg,
                                 mime="text/plain;charset=UTF-8")

            helpmsg = "Rerun all other cases defined in the case selector."
            col2.button("Rerun all cases", on_click=owb.runAllCases, help=helpmsg)

        lines = kz.getCaseKey("casetoml")
        if lines != "":
            st.divider()
            st.write("#### :orange[Case Parameter File]\n"
                     "This file contains the parameters characterizing the current case"
                     " and can be used, along with the *Wages and Contributions* file, to reproduce it in the future.")
            st.code(lines, height=400, language="toml")

            st.download_button(
                "Download case parameter file", data=lines,
                file_name=f"case_{caseName}.toml", mime="application/toml"
            )

        st.divider()
        st.write("#### :orange[Excel Workbooks]\n"
                 "These workbooks contain time tables describing the flow of money,"
                 " the first one as input to the case, and the second as its output.")
        col1, col2 = st.columns(2, gap="large")
        with col1:
            download2 = st.download_button(
                label="Download Wages and Contributions file",
                help="Download Wages and Contributions as an Excel workbook.",
                data=owb.saveContributions(),
                file_name=f"{caseName}.xlsx",
                disabled=kz.isCaseUnsolved(),
                mime="application/vnd.ms-excel",
            )

        with col2:
            download2 = st.download_button(
                label="Download Worksheets",
                help="Download Worksheets as an Excel workbook.",
                data=owb.saveWorkbook(),
                file_name=f"Workbook_{caseName}.xlsx",
                mime="application/vnd.ms-excel",
                disabled=kz.isCaseUnsolved(),
            )
